    package com.croftsoft.apps.tpu;
    
    import java.util.*;
    import java.util.concurrent.*;

    /***********************************************************************
    * Three phase update demo implementation.
    * 
    * For a tutorial on the use of the Three Phase Update (TPU) pattern,
    * please see Croft, "Concurrent Java Simulations Using Three Phase
    * Update", 2009-07-04, <a target="blank"
    * href="http://www.CroftSoft.com/library/tutorials/tpu/">
    *       http://www.CroftSoft.com/library/tutorials/tpu/</a>.
    *       
    * @version
    *   $Id: ThreePhaseUpdateCell.java 27 2009-07-04 06:37:43Z croft $
    * @since
    *   2009-05-10
    * @author
    *   <a href="http://www.CroftSoft.com/">David Wallace Croft</a>
    ***********************************************************************/

    public final class  ThreePhaseUpdateCell
      implements ThreePhaseUpdate
    ////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////
    {
      
    private final String  id;
      
    private final Random  random; 
      
    // All non-final instance variables must be volatile.      

    /** Set after construction and referenced during access(). */
    private volatile ThreePhaseUpdateCell [ ]  adjacentCells;
    
    /** The current state. */
    private volatile boolean  alive;
    
    /** Calculated during access() and used in digest(). */
    private volatile int  adjacentCellsAliveCount;
    
    /** This is the next state which is set when mutate() is called. */
    private volatile boolean  nextAlive;
    
    ////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////

    public  ThreePhaseUpdateCell (
      final String  id,
      final Random  random )
    ////////////////////////////////////////////////////////////////////////
    {
      this.id     = id;
      
      this.random = random;
    }
    
    ////////////////////////////////////////////////////////////////////////
    // accessor methods
    ////////////////////////////////////////////////////////////////////////
    
    public String   getId   ( ) { return id;    }
    
    public boolean  isAlive ( ) { return alive; }
      
    ////////////////////////////////////////////////////////////////////////
    // mutator methods
    ////////////////////////////////////////////////////////////////////////
    
    public void  setAlive ( final boolean  alive )
    ////////////////////////////////////////////////////////////////////////
    {
      this.alive = alive;
    }
    
    public void  setAdjacentCells (
      final ThreePhaseUpdateCell [ ]  adjacentCells )
    ////////////////////////////////////////////////////////////////////////
    {
      this.adjacentCells = adjacentCells;
    }
    
    ////////////////////////////////////////////////////////////////////////
    // interface ThreePhaseUpdate methods
    ////////////////////////////////////////////////////////////////////////
      
    @Override
    public void  access ( )
    ////////////////////////////////////////////////////////////////////////
    {
      System.out.println (
        id + ":  access begin " + Thread.currentThread ( ) );
      
      adjacentCellsAliveCount = 0;
      
      for ( final ThreePhaseUpdateCell  adjacentCell : adjacentCells )
      {
        if ( adjacentCell.isAlive ( ) )
        {
          adjacentCellsAliveCount++;
        }
      }
      
      randomDelay ( 1 );
          
      System.out.println (
        id + ":  access done  " + Thread.currentThread ( ) );
    }
    
    @Override
    public void  digest ( )
    ////////////////////////////////////////////////////////////////////////
    {
      System.out.println (
        id + ":  digest begin " + Thread.currentThread ( ) );
      
      nextAlive = alive;
      
      if ( alive )
      {
        if ( ( adjacentCellsAliveCount < 2 )
          || ( adjacentCellsAliveCount > 3 ) )
        {
          nextAlive = false;
        }
      }
      else
      {
        if ( adjacentCellsAliveCount == 3 )
        {
          nextAlive = true;
        }
      }
      
      randomDelay ( 1000 );
      
      System.out.println (
        id + ":  digest done  " + Thread.currentThread ( ) );
    }
    
    @Override
    public void  mutate ( )
    ////////////////////////////////////////////////////////////////////////
    {
      System.out.println (
        id + ":  mutate begin " + Thread.currentThread ( ) );
      
      alive = nextAlive;
      
      randomDelay ( 1 );
      
      System.out.println (
        id + ":  mutate done  " + Thread.currentThread ( ) );
    }
    
    ////////////////////////////////////////////////////////////////////////
    // private methods
    ////////////////////////////////////////////////////////////////////////
    
    private void  randomDelay ( final int  maxDelay )
    ////////////////////////////////////////////////////////////////////////
    {
      try
      {      
        TimeUnit.MILLISECONDS.sleep ( random.nextInt ( maxDelay ) );
      }
      catch ( final InterruptedException  ie )
      {
        // ignore
      }
    }
    
    ////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////
    }